@@ -4,13 +4,14 @@
 
 from .sortedset import SortedSet
 from .sortedlist import SortedList, recursive_repr
-from collections import Mapping, MutableMapping, Set, Sequence
+from .sortedlistwithkey import SortedListWithKey
+from collections import Set, Sequence
 from collections import KeysView as AbstractKeysView
 from collections import ValuesView as AbstractValuesView
 from collections import ItemsView as AbstractItemsView
 
 from functools import wraps
-from sys import version_info, hexversion
+from sys import hexversion
 
 _NotGiven = object()
 
@@ -44,19 +45,20 @@ class _IlocWrapper:
         indices and slice notation. Raises IndexError on invalid *index*.
         """
         _temp = self._dict
-        _list, _dict = _temp._list, _temp._dict
+        _list = _temp._list
+        _delitem = _temp._delitem
 
         if isinstance(index, slice):
             keys = _list[index]
             del _list[index]
             for key in keys:
-                del _dict[key]
+                _delitem(key)
         else:
             key = _list[index]
             del _list[index]
-            del _dict[key]
+            _delitem(key)
 
-class SortedDict(MutableMapping):
+class SortedDict(dict):
     """
     A SortedDict provides the same methods as a dict.  Additionally, a
     SortedDict efficiently maintains its keys in sorted order. Consequently, the
@@ -70,13 +72,19 @@ class SortedDict(MutableMapping):
         the keys method will return the keys in sorted order, the popitem method
         will remove the item with the highest key, etc.
 
+        An optional *key* argument defines a callable that, like the `key`
+        argument to Python's `sorted` function, extracts a comparison key from
+        each dict key. If no function is specified, the default compares the
+        dict keys directly. The `key` argument must be provided as a positional
+        argument and must come before all other arguments.
+
         An optional *load* argument defines the load factor of the internal list
-        used to maintain sort order. If present, this argument must come first
-        and must be an integer. The default load factor of '1000' works well for
-        lists from tens to tens of millions of elements.  Good practice is to
-        use a value that is the cube root of the list size.  With billions of
-        elements, the best load factor depends on your usage.  It's best to
-        leave the load factor at the default until you start benchmarking.
+        used to maintain sort order. If present, this argument must come before
+        an iterable. The default load factor of '1000' works well for lists from
+        tens to tens of millions of elements.  Good practice is to use a value
+        that is the cube root of the list size.  With billions of elements, the
+        best load factor depends on your usage.  It's best to leave the load
+        factor at the default until you start benchmarking.
 
         An optional *iterable* argument provides an initial series of items to
         populate the SortedDict.  Each item in the series must itself contain
@@ -99,53 +107,72 @@ class SortedDict(MutableMapping):
         The first example only works for keys that are valid Python
         identifiers; the others work with any valid keys.
         """
-        if len(args) > 0 and type(args[0]) == int:
-            load = args[0]
+        if len(args) > 0 and (args[0] is None or callable(args[0])):
+            self._key = args[0]
             args = args[1:]
         else:
-            load = 1000
+            self._key = None
+
+        if len(args) > 0 and type(args[0]) == int:
+            self._load = args[0]
+            args = args[1:]
+        else:
+            self._load = 1000
+
+        if self._key is None:
+            self._list = SortedList(load=self._load)
+        else:
+            self._list = SortedListWithKey(key=self._key, load=self._load)
+
+        # Cache function pointers to dict methods.
+
+        _dict = super(SortedDict, self)
+        self._dict = _dict
+        self._clear = _dict.clear
+        self._delitem = _dict.__delitem__
+        self._iter = _dict.__iter__
+        self._pop = _dict.pop
+        self._setdefault = _dict.setdefault
+        self._setitem = _dict.__setitem__
+        self._update = _dict.update
+
+        # Cache function pointers to SortedList methods.
+
+        _list = self._list
+        self._list_add = _list.add
+        self.bisect_left = _list.bisect_left
+        self.bisect = _list.bisect_right
+        self.bisect_right = _list.bisect_right
+        self._list_clear = _list.clear
+        self.index = _list.index
+        self._list_pop = _list.pop
+        self._list_remove = _list.remove
+        self._list_update = _list.update
+        self.irange = _list.irange
+        self.islice = _list.islice
+
+        if self._key is not None:
+            self.bisect_key_left = _list.bisect_key_left
+            self.bisect_key_right = _list.bisect_key_right
+            self.bisect_key = _list.bisect_key
+            self.irange_key = _list.irange_key
 
-        self._dict = dict()
-        self._list = SortedList(load=load)
         self.iloc = _IlocWrapper(self)
 
         self.update(*args, **kwargs)
 
     def clear(self):
         """Remove all elements from the dictionary."""
-        self._dict.clear()
-        self._list.clear()
-
-    def __contains__(self, key):
-        """Return True if and only if *key* is in the dictionary."""
-        return key in self._dict
+        self._clear()
+        self._list_clear()
 
     def __delitem__(self, key):
         """
         Remove ``d[key]`` from *d*.  Raises a KeyError if *key* is not in the
         dictionary.
         """
-        del self._dict[key]
-        self._list.remove(key)
-
-    def __getitem__(self, key):
-        """
-        Return the item of *d* with key *key*.  Raises a KeyError if *key* is
-        not in the dictionary.
-        """
-        return self._dict[key]
-
-    def __eq__(self, that):
-        """Compare two iterables for equality."""
-        return (len(self._dict) == len(that)
-                and all((key in that) and (self[key] == that[key])
-                        for key in self))
-
-    def __ne__(self, that):
-        """Compare two iterables for inequality."""
-        return (len(self._dict) != len(that)
-                or any((key not in that) or (self[key] != that[key])
-                       for key in self))
+        self._delitem(key)
+        self._list_remove(key)
 
     def __iter__(self):
         """Create an iterator over the sorted keys of the dictionary."""
@@ -157,98 +184,77 @@ class SortedDict(MutableMapping):
         """
         return reversed(self._list)
 
-    def __len__(self):
-        """Return the number of (key, value) pairs in the dictionary."""
-        return len(self._dict)
-
     def __setitem__(self, key, value):
         """Set `d[key]` to *value*."""
-        _dict = self._dict
-        if key not in _dict:
-            self._list.add(key)
-        _dict[key] = value
+        if key not in self:
+            self._list_add(key)
+        self._setitem(key, value)
 
     def copy(self):
         """Return a shallow copy of the sorted dictionary."""
-        return SortedDict(self._list._load, self._dict)
+        return self.__class__(self._key, self._load, self.iteritems())
 
-    def __copy__(self):
-        """Return a shallow copy of the sorted dictionary."""
-        return self.copy()
+    __copy__ = copy
 
     @classmethod
     def fromkeys(cls, seq, value=None):
         """
         Create a new dictionary with keys from *seq* and values set to *value*.
         """
-        that = SortedDict((key, value) for key in seq)
-        return that
+        return cls((key, value) for key in seq)
 
-    def get(self, key, default=None):
-        """
-        Return the value for *key* if *key* is in the dictionary, else
-        *default*.  If *default* is not given, it defaults to ``None``,
-        so that this method never raises a KeyError.
-        """
-        return self._dict.get(key, default)
-
-    def has_key(self, key):
-        """Return True if and only in *key* is in the dictionary."""
-        return key in self._dict
-
-    def items(self):
-        """
-        In Python 2, returns a list of the dictionary's items (``(key, value)``
-        pairs).
-
-        In Python 3, returns a new ItemsView of the dictionary's items.  In
-        addition to the methods provided by the built-in `view` the ItemsView is
-        indexable (e.g., ``d.items()[5]``).
-        """
-        if version_info[0] == 2:
+    if hexversion < 0x03000000:
+        def items(self):
+            """
+            Return a list of the dictionary's items (``(key, value)`` pairs).
+            """
             return list(self.iteritems())
-        else:
+    else:
+        def items(self):
+            """
+            Return a new ItemsView of the dictionary's items.  In addition to
+            the methods provided by the built-in `view` the ItemsView is
+            indexable (e.g. ``d.items()[5]``).
+            """
             return ItemsView(self)
 
     def iteritems(self):
         """Return an iterable over the items (``(key, value)`` pairs)."""
-        _dict = self._dict
-        return iter((key, _dict[key]) for key in self._list)
+        return iter((key, self[key]) for key in self._list)
 
-    def keys(self):
-        """
-        In Python 2, return a SortedSet of the dictionary's keys.
-
-        In Python 3, return a new KeysView of the dictionary's keys.  In
-        addition to the methods provided by the built-in `view` the KeysView is
-        indexable (e.g., ``d.keys()[5]``).
-        """
-        if version_info[0] == 2:
-            return SortedSet(self._dict)
-        else:
+    if hexversion < 0x03000000:
+        def keys(self):
+            """Return a SortedSet of the dictionary's keys."""
+            return SortedSet(self._list, key=self._key, load=self._load)
+    else:
+        def keys(self):
+            """
+            Return a new KeysView of the dictionary's keys.  In addition to the
+            methods provided by the built-in `view` the KeysView is indexable
+            (e.g. ``d.keys()[5]``).
+            """
             return KeysView(self)
 
     def iterkeys(self):
         """Return an iterable over the keys of the dictionary."""
         return iter(self._list)
 
-    def values(self):
-        """
-        In Python 2, return a list of the dictionary's values.
-
-        In Python 3, return a new :class:`ValuesView` of the dictionary's
-        values.  In addition to the methods provided by the built-in `view`
-        the ValuesView is indexable (e.g., ``d.values()[5]``).
-        """
-        if version_info[0] == 2:
+    if hexversion < 0x03000000:
+        def values(self):
+            """Return a list of the dictionary's values."""
             return list(self.itervalues())
-        else:
+    else:
+        def values(self):
+            """
+            Return a new :class:`ValuesView` of the dictionary's values.
+            In addition to the methods provided by the built-in `view` the
+            ValuesView is indexable (e.g., ``d.values()[5]``).
+            """
             return ValuesView(self)
 
     def itervalues(self):
         """Return an iterable over the values of the dictionary."""
-        _dict = self._dict
-        return iter(_dict[key] for key in self._list)
+        return iter(self[key] for key in self._list)
 
     def pop(self, key, default=_NotGiven):
         """
@@ -256,31 +262,29 @@ class SortedDict(MutableMapping):
         else return *default*. If *default* is not given and *key* is not in
         the dictionary, a KeyError is raised.
         """
-        if key in self._dict:
-            self._list.remove(key)
-            return self._dict.pop(key)
+        if key in self:
+            self._list_remove(key)
+            return self._pop(key)
         else:
-            if default == _NotGiven:
-                raise KeyError
+            if default is _NotGiven:
+                raise KeyError(key)
             else:
                 return default
 
-    def popitem(self):
+    def popitem(self, last=True):
         """
-        Remove and return the ``(key, value)`` pair with the greatest *key*
-        from the dictionary.
+        Remove and return a ``(key, value)`` pair from the dictionary. If
+        last=True (default) then remove the *greatest* `key` from the
+        diciontary. Else, remove the *least* key from the dictionary.
 
         If the dictionary is empty, calling `popitem` raises a
         KeyError`.
         """
-        _dict, _list = self._dict, self._list
+        if not len(self):
+            raise KeyError('popitem(): dictionary is empty')
 
-        if len(_dict) == 0:
-            raise KeyError
-
-        key = _list.pop()
-        value = _dict[key]
-        del _dict[key]
+        key = self._list_pop(-1 if last else 0)
+        value = self._pop(key)
 
         return (key, value)
 
@@ -290,12 +294,11 @@ class SortedDict(MutableMapping):
         with a value of *default* and return *default*.  *default* defaults to
         ``None``.
         """
-        _dict = self._dict
-        if key in _dict:
-            return _dict[key]
+        if key in self:
+            return self[key]
         else:
-            _dict[key] = default
-            self._list.add(key)
+            self._setitem(key, default)
+            self._list_add(key)
             return default
 
     def update(self, *args, **kwargs):
@@ -308,11 +311,9 @@ class SortedDict(MutableMapping):
         keyword arguments are specified, the dictionary is then updated with
         those key/value pairs: ``d.update(red=1, blue=2)``.
         """
-        _dict, _list = self._dict, self._list
-
-        if len(_dict) == 0:
-            _dict.update(*args, **kwargs)
-            _list.update(_dict)
+        if not len(self):
+            self._update(*args, **kwargs)
+            self._list_update(self._iter())
             return
 
         if (len(kwargs) == 0 and len(args) == 1 and isinstance(args[0], dict)):
@@ -320,45 +321,14 @@ class SortedDict(MutableMapping):
         else:
             pairs = dict(*args, **kwargs)
 
-        if (10 * len(pairs)) > len(self._dict):
-            self._dict.update(pairs)
-            _list = self._list
-            _list.clear()
-            _list.update(self._dict)
+        if (10 * len(pairs)) > len(self):
+            self._update(pairs)
+            self._list_clear()
+            self._list_update(self._iter())
         else:
             for key in pairs:
                 self[key] = pairs[key]
 
-    def index(self, key, start=None, stop=None):
-        """
-        Return the smallest *k* such that `d.iloc[k] == key` and `i <= k < j`.
-        Raises `ValueError` if *key* is not present.  *stop* defaults to the end
-        of the set.  *start* defaults to the beginning.  Negative indexes are
-        supported, as for slice indices.
-        """
-        return self._list.index(key, start, stop)
-
-    def bisect_left(self, key):
-        """
-        Similar to the ``bisect`` module in the standard library, this returns
-        an appropriate index to insert *key* in SortedDict. If *key* is
-        already present in SortedDict, the insertion point will be before (to the
-        left of) any existing entries.
-        """
-        return self._list.bisect_left(key)
-
-    def bisect(self, key):
-        """Same as bisect_left."""
-        return self._list.bisect(key)
-
-    def bisect_right(self, key):
-        """
-        Same as `bisect_left`, but if *key* is already present in SortedDict,
-        the insertion point will be after (to the right of) any existing
-        entries.
-        """
-        return self._list.bisect_right(key)
-
     @not26
     def viewkeys(self):
         """
@@ -389,17 +359,25 @@ class SortedDict(MutableMapping):
         """
         return ItemsView(self)
 
+    def __reduce__(self):
+        return (self.__class__, (self._key, self._load, list(self.iteritems())))
+
     @recursive_repr
     def __repr__(self):
-        _dict = self._dict
-        items = ', '.join('{0}: {1}'.format(repr(key), repr(_dict[key]))
+        temp = '{0}({1}, {2}, {{{3}}})'
+        items = ', '.join('{0}: {1}'.format(repr(key), repr(self[key]))
                           for key in self._list)
-        return '{0}({{{1}}})'.format(self.__class__.__name__, items)
+        return temp.format(
+            self.__class__.__name__,
+            repr(self._key),
+            repr(self._load),
+            items
+        )
 
     def _check(self):
         self._list._check()
-        assert len(self._dict) == len(self._list)
-        assert all(val in self._dict for val in self._list)
+        assert len(self) == len(self._list)
+        assert all(val in self for val in self._list)
 
 class KeysView(AbstractKeysView, Set, Sequence):
     """
@@ -409,14 +387,19 @@ class KeysView(AbstractKeysView, Set, Sequence):
 
     The KeysView class implements the Set and Sequence Abstract Base Classes.
     """
-    def __init__(self, sorted_dict):
-        """
-        Initialize a KeysView from a SortedDict container as *sorted_dict*.
-        """
-        self._list = sorted_dict._list
-        if version_info[0] == 2:
+    if hexversion < 0x03000000:
+        def __init__(self, sorted_dict):
+            """
+            Initialize a KeysView from a SortedDict container as *sorted_dict*.
+            """
+            self._list = sorted_dict._list
             self._view = sorted_dict._dict.viewkeys()
-        else:
+    else:
+        def __init__(self, sorted_dict):
+            """
+            Initialize a KeysView from a SortedDict container as *sorted_dict*.
+            """
+            self._list = sorted_dict._list
             self._view = sorted_dict._dict.keys()
     def __len__(self):
         """Return the number of entries in the dictionary."""
@@ -489,11 +472,13 @@ class KeysView(AbstractKeysView, Set, Sequence):
     def __xor__(self, that):
         """Return a SortedSet of the symmetric difference of self and *that*."""
         return SortedSet(self._view ^ that)
-    def isdisjoint(self, that):
-        """Return True if and only if *that* is disjoint with self."""
-        if version_info[0] == 2:
+    if hexversion < 0x03000000:
+        def isdisjoint(self, that):
+            """Return True if and only if *that* is disjoint with self."""
             return not any(key in self._list for key in that)
-        else:
+    else:
+        def isdisjoint(self, that):
+            """Return True if and only if *that* is disjoint with self."""
             return self._view.isdisjoint(that)
     @recursive_repr
     def __repr__(self):
@@ -507,15 +492,23 @@ class ValuesView(AbstractValuesView, Sequence):
 
     The ValuesView class implements the Sequence Abstract Base Class.
     """
-    def __init__(self, sorted_dict):
-        """
-        Initialize a ValuesView from a SortedDict container as *sorted_dict*.
-        """
-        self._dict = sorted_dict
-        self._list = sorted_dict._list
-        if version_info[0] == 2:
+    if hexversion < 0x03000000:
+        def __init__(self, sorted_dict):
+            """
+            Initialize a ValuesView from a SortedDict container as
+            *sorted_dict*.
+            """
+            self._dict = sorted_dict
+            self._list = sorted_dict._list
             self._view = sorted_dict._dict.viewvalues()
-        else:
+    else:
+        def __init__(self, sorted_dict):
+            """
+            Initialize a ValuesView from a SortedDict container as
+            *sorted_dict*.
+            """
+            self._dict = sorted_dict
+            self._list = sorted_dict._list
             self._view = sorted_dict._dict.values()
     def __len__(self):
         """Return the number of entries in the dictionary."""
@@ -567,13 +560,14 @@ class ValuesView(AbstractValuesView, Sequence):
             if value == val:
                 return idx
         else:
-            raise ValueError
-    def count(self, value):
-        """Return the number of occurrences of *value* in self."""
-        _dict = self._dict
-        if version_info[0] == 2:
-            return sum(1 for val in _dict.itervalues() if val == value)
-        else:
+            raise ValueError('{0} is not in dict'.format(repr(value)))
+    if hexversion < 0x03000000:
+        def count(self, value):
+            """Return the number of occurrences of *value* in self."""
+            return sum(1 for val in self._dict.itervalues() if val == value)
+    else:
+        def count(self, value):
+            """Return the number of occurrences of *value* in self."""
             return sum(1 for val in _dict.values() if val == value)
     def __lt__(self, that):
         raise TypeError
@@ -605,15 +599,23 @@ class ItemsView(AbstractItemsView, Set, Sequence):
     However, the set-like operations (``&``, ``|``, ``-``, ``^``) will only
     operate correctly if all of the dictionary's values are hashable.
     """
-    def __init__(self, sorted_dict):
-        """
-        Initialize an ItemsView from a SortedDict container as *sorted_dict*.
-        """
-        self._dict = sorted_dict
-        self._list = sorted_dict._list
-        if version_info[0] == 2:
+    if hexversion < 0x03000000:
+        def __init__(self, sorted_dict):
+            """
+            Initialize an ItemsView from a SortedDict container as
+            *sorted_dict*.
+            """
+            self._dict = sorted_dict
+            self._list = sorted_dict._list
             self._view = sorted_dict._dict.viewitems()
-        else:
+    else:
+        def __init__(self, sorted_dict):
+            """
+            Initialize an ItemsView from a SortedDict container as
+            *sorted_dict*.
+            """
+            self._dict = sorted_dict
+            self._list = sorted_dict._list
             self._view = sorted_dict._dict.items()
     def __len__(self):
         """Return the number of entries in the dictionary."""
@@ -659,11 +661,12 @@ class ItemsView(AbstractItemsView, Set, Sequence):
         to the end of the set.  *start* defaults to the beginning.  Negative
         indexes are supported, as for slice indices.
         """
-        pos = self._list.index(key[0], start, stop)
-        if key[1] == self._dict[key[0]]:
+        temp, value = key
+        pos = self._list.index(temp, start, stop)
+        if value == self._dict[temp]:
             return pos
         else:
-            raise ValueError
+            raise ValueError('{0} is not in dict'.format(repr(key)))
     def count(self, item):
         """Return the number of occurrences of *item* in the set."""
         key, value = item
@@ -698,15 +701,17 @@ class ItemsView(AbstractItemsView, Set, Sequence):
     def __xor__(self, that):
         """Return a SortedSet of the symmetric difference of self and *that*."""
         return SortedSet(self._view ^ that)
-    def isdisjoint(self, that):
-        """Return True if and only if *that* is disjoint with self."""
-        if version_info[0] == 2:
+    if hexversion < 0x03000000:
+        def isdisjoint(self, that):
+            """Return True if and only if *that* is disjoint with self."""
             _dict = self._dict
             for key, value in that:
                 if key in _dict and _dict[key] == value:
                     return False
             return True
-        else:
+    else:
+        def isdisjoint(self, that):
+            """Return True if and only if *that* is disjoint with self."""
             return self._view.isdisjoint(that)
     @recursive_repr
     def __repr__(self):
