@@ -69,6 +69,7 @@ class SortedList(MutableSequence):
         """
         self._len, self._maxes, self._lists, self._index = 0, [], [], []
         self._load, self._twice, self._half = load, load * 2, load >> 1
+        self._offset = 0
 
         if iterable is not None:
             self.update(iterable)
@@ -113,7 +114,7 @@ class SortedList(MutableSequence):
         if len(_lists[pos]) > self._twice:
             _maxes, _load = self._maxes, self._load
             half = _lists[pos][_load:]
-            _lists[pos] = _lists[pos][:_load]
+            del _lists[pos][_load:]
             _maxes[pos] = _lists[pos][-1]
             _maxes.insert(pos + 1, half[-1])
             _lists.insert(pos + 1, half)
@@ -144,7 +145,7 @@ class SortedList(MutableSequence):
 
         _load, _index = self._load, self._index
         _lists.extend(values[pos:(pos + _load)]
-                           for pos in range(0, len(values), _load))
+                      for pos in range(0, len(values), _load))
         _maxes.extend(sublist[-1] for sublist in _lists)
         self._len = len(values)
         del _index[:]
@@ -195,19 +196,19 @@ class SortedList(MutableSequence):
         _maxes = self._maxes
 
         if not _maxes:
-            raise ValueError
+            raise ValueError('{0} not in list'.format(repr(val)))
 
         pos = bisect_left(_maxes, val)
 
         if pos == len(_maxes):
-            raise ValueError
+            raise ValueError('{0} not in list'.format(repr(val)))
 
         _lists = self._lists
         idx = bisect_left(_lists[pos], val)
         if _lists[pos][idx] == val:
             self._delete(pos, idx)
         else:
-            raise ValueError
+            raise ValueError('{0} not in list'.format(repr(val)))
 
     def _delete(self, pos, idx):
         """Delete the item at the given (pos, idx).
@@ -240,7 +241,8 @@ class SortedList(MutableSequence):
 
         elif len(_lists) > 1:
 
-            if pos == 0: pos += 1
+            if not pos:
+                pos += 1
 
             prev = pos - 1
             _lists[prev].extend(_lists[pos])
@@ -269,11 +271,11 @@ class SortedList(MutableSequence):
         Most queries require the index be built. Details of the index are
         described in self._build_index.
 
-        Indexing requires traversing the tree from a leaf node to the root. The parent
-        of each node is easily computable at (pos - 1) // 2.
+        Indexing requires traversing the tree from a leaf node to the root. The
+        parent of each node is easily computable at (pos - 1) // 2.
 
-        Left-child nodes are always at odd indices and right-child nodes are always
-        at even indices.
+        Left-child nodes are always at odd indices and right-child nodes are
+        always at even indices.
 
         When traversing up from a right-child node, increment the total by the
         left-child node.
@@ -306,12 +308,12 @@ class SortedList(MutableSequence):
 
         Computing the index is the sum of the total and beta: 5 + 3 = 8.
         """
-        if pos == 0:
+        if not pos:
             return idx
 
         _index = self._index
 
-        if len(_index) == 0:
+        if not len(_index):
             self._build_index()
 
         total = 0
@@ -386,29 +388,27 @@ class SortedList(MutableSequence):
         The final index pair from our example is (2, 3) which corresponds to
         index 8 in the sorted list.
         """
-        _len, _lists = self._len, self._lists
-
         if idx < 0:
-            last_len = len(_lists[-1])
+            last_len = len(self._lists[-1])
             if (-idx) <= last_len:
-                return len(_lists) - 1, last_len + idx
-            idx += _len
+                return len(self._lists) - 1, last_len + idx
+            idx += self._len
             if idx < 0:
-                raise IndexError
-        elif idx >= _len:
-            raise IndexError
+                raise IndexError('list index out of range')
+        elif idx >= self._len:
+            raise IndexError('list index out of range')
 
-        if idx < len(_lists[0]):
+        if idx < len(self._lists[0]):
             return 0, idx
 
         _index = self._index
 
-        if len(_index) == 0:
+        if not _index:
             self._build_index()
 
         pos = 0
+        child = 1
         len_index = len(_index)
-        child = (pos << 1) + 1
 
         while child < len_index:
             index_child = _index[child]
@@ -553,7 +553,7 @@ class SortedList(MutableSequence):
             start, stop, step = self._slice(idx)
 
             if ((step == 1) and (start < stop)
-                and ((stop - start) * 8 >= self._len)):
+                    and ((stop - start) * 8 >= self._len)):
 
                 values = self[:start]
                 if stop < self._len:
@@ -628,7 +628,8 @@ class SortedList(MutableSequence):
 
         pos, loc = self._pos(idx)
 
-        if idx < 0: idx += _len
+        if idx < 0:
+            idx += _len
 
         # Check that the inserted value is not less than the
         # previous value.
@@ -642,7 +643,8 @@ class SortedList(MutableSequence):
                 idx_prev = len(_lists[pos_prev]) - 1
 
             if _lists[pos_prev][idx_prev] > val:
-                raise ValueError
+                msg = '{0} not in sort order at index {1}'.format(repr(val), idx)
+                raise ValueError(msg)
 
         # Check that the inserted value is not greater than
         # the previous value.
@@ -656,7 +658,8 @@ class SortedList(MutableSequence):
                 idx_next = 0
 
             if _lists[pos_next][idx_next] < val:
-                raise ValueError
+                msg = '{0} not in sort order at index {1}'.format(repr(val), idx)
+                raise ValueError(msg)
 
     def __setitem__(self, index, value):
         """
@@ -667,7 +670,8 @@ class SortedList(MutableSequence):
         :exc:`ValueError` is raised before the list is mutated if the sort order
         would be violated by the operation.
         """
-        _pos, _lists, _check_order = self._pos, self._lists, self._check_order
+        _maxes, _lists, _pos = self._maxes, self._lists, self._pos
+        _check_order = self._check_order
 
         if isinstance(index, slice):
             start, stop, step = self._slice(index)
@@ -695,6 +699,8 @@ class SortedList(MutableSequence):
                     pos, loc = _pos(idx)
                     _append((idx, _lists[pos][loc], val))
                     _lists[pos][loc] = val
+                    if len(_lists[pos]) == (loc + 1):
+                        _maxes[pos] = val
 
                 try:
                     # Validate ordering of new values.
@@ -709,6 +715,8 @@ class SortedList(MutableSequence):
                     for idx, oldval, newval in log:
                         pos, loc = _pos(idx)
                         _lists[pos][loc] = oldval
+                        if len(_lists[pos]) == (loc + 1):
+                            _maxes[pos] = oldval
 
                     raise
             else:
@@ -724,25 +732,27 @@ class SortedList(MutableSequence):
                               for pos in range(1, len(value)))
 
                 if not ordered:
-                    raise ValueError
+                    raise ValueError('given sequence not in sort order')
 
                 # Check ordering in context of sorted list.
 
-                if start == 0 or len(value) == 0:
+                if not start or not len(value):
                     # Nothing to check on the lhs.
                     pass
                 else:
                     if self[start - 1] > value[0]:
-                        raise ValueError
+                        msg = '{0} not in sort order at index {1}'.format(repr(value[0]), start)
+                        raise ValueError(msg)
 
-                if stop == len(self) or len(value) == 0:
+                if stop == len(self) or not len(value):
                     # Nothing to check on the rhs.
                     pass
                 else:
                     # "stop" is exclusive so we don't need
                     # to add one for the index.
                     if self[stop] < value[-1]:
-                        raise ValueError
+                        msg = '{0} not in sort order at index {1}'.format(repr(value[-1]), stop)
+                        raise ValueError(msg)
 
                 # Delete the existing values.
 
@@ -757,6 +767,8 @@ class SortedList(MutableSequence):
             pos, loc = _pos(index)
             _check_order(index, value)
             _lists[pos][loc] = value
+            if len(_lists[pos]) == (loc + 1):
+                _maxes[pos] = value
 
     def __iter__(self):
         """Create an iterator over the list."""
@@ -764,11 +776,150 @@ class SortedList(MutableSequence):
 
     def __reversed__(self):
         """Create an iterator to traverse the list in reverse."""
+        return chain.from_iterable(map(reversed, reversed(self._lists)))
+
+    def islice(self, start=None, stop=None, reverse=False):
+        """
+        Returns an iterator that slices `self` from `start` to `stop` index,
+        inclusive and exclusive respectively.
+
+        When `reverse` is `True`, values are yielded from the iterator in
+        reverse order.
+
+        Both `start` and `stop` default to `None` which is automatically
+        inclusive of the beginning and end.
+        """
+        _len = self._len
+
+        if not _len:
+            return iter(())
+
+        start, stop, step = self._slice(slice(start, stop))
+
+        if start >= stop:
+            return iter(())
+
+        _pos = self._pos
+
+        min_pos, min_idx = _pos(start)
+
+        if stop == _len:
+            max_pos = len(self._lists) - 1
+            max_idx = len(self._lists[-1])
+        else:
+            max_pos, max_idx = _pos(stop)
+
+        return self._islice(min_pos, min_idx, max_pos, max_idx, reverse)
+
+    def _islice(self, min_pos, min_idx, max_pos, max_idx, reverse):
+        """
+        Returns an iterator that slices `self` using two index pairs,
+        `(min_pos, min_idx)` and `(max_pos, max_idx)`; the first inclusive
+        and the latter exclusive. See `_pos` for details on how an index
+        is converted to an index pair.
+
+        When `reverse` is `True`, values are yielded from the iterator in
+        reverse order.
+        """
         _lists = self._lists
-        start = len(_lists) - 1
-        iterable = (reversed(_lists[pos])
-                    for pos in range(start, -1, -1))
-        return chain.from_iterable(iterable)
+
+        if min_pos > max_pos:
+            return iter(())
+        elif min_pos == max_pos and not reverse:
+            return iter(_lists[min_pos][min_idx:max_idx])
+        elif min_pos == max_pos and reverse:
+            return reversed(_lists[min_pos][min_idx:max_idx])
+        elif min_pos + 1 == max_pos and not reverse:
+            return chain(_lists[min_pos][min_idx:], _lists[max_pos][:max_idx])
+        elif min_pos + 1 == max_pos and reverse:
+            return chain(
+                reversed(_lists[max_pos][:max_idx]),
+                reversed(_lists[min_pos][min_idx:]),
+            )
+        elif not reverse:
+            return chain(
+                _lists[min_pos][min_idx:],
+                chain.from_iterable(_lists[(min_pos + 1):max_pos]),
+                _lists[max_pos][:max_idx],
+            )
+        else:
+            temp = map(reversed, reversed(_lists[(min_pos + 1):max_pos]))
+            return chain(
+                reversed(_lists[max_pos][:max_idx]),
+                chain.from_iterable(temp),
+                reversed(_lists[min_pos][min_idx:]),
+            )
+
+    def irange(self, minimum=None, maximum=None, inclusive=(True, True),
+               reverse=False):
+        """
+        Create an iterator of values between `minimum` and `maximum`.
+
+        `inclusive` is a pair of booleans that indicates whether the minimum
+        and maximum ought to be included in the range, respectively. The
+        default is (True, True) such that the range is inclusive of both
+        minimum and maximum.
+
+        Both `minimum` and `maximum` default to `None` which is automatically
+        inclusive of the start and end of the list, respectively.
+
+        When `reverse` is `True` the values are yielded from the iterator in
+        reverse order; `reverse` defaults to `False`.
+        """
+        _maxes = self._maxes
+
+        if not _maxes:
+            return iter(())
+
+        _lists = self._lists
+
+        # Calculate the minimum (pos, idx) pair. By default this location
+        # will be inclusive in our calculation.
+
+        if minimum is None:
+            min_pos = 0
+            min_idx = 0
+        else:
+            if inclusive[0]:
+                min_pos = bisect_left(_maxes, minimum)
+
+                if min_pos == len(_maxes):
+                    return iter(())
+
+                min_idx = bisect_left(_lists[min_pos], minimum)
+            else:
+                min_pos = bisect_right(_maxes, minimum)
+
+                if min_pos == len(_maxes):
+                    return iter(())
+
+                min_idx = bisect_right(_lists[min_pos], minimum)
+
+        # Calculate the maximum (pos, idx) pair. By default this location
+        # will be exclusive in our calculation.
+
+        if maximum is None:
+            max_pos = len(_maxes) - 1
+            max_idx = len(_lists[max_pos])
+        else:
+            if inclusive[1]:
+                max_pos = bisect_right(_maxes, maximum)
+
+                if max_pos == len(_maxes):
+                    max_pos -= 1
+                    max_idx = len(_lists[max_pos])
+                else:
+                    max_idx = bisect_right(_lists[max_pos], maximum)
+            else:
+                max_pos = bisect_left(_maxes, maximum)
+
+                if max_pos == len(_maxes):
+                    max_pos -= 1
+                    max_idx = len(_lists[max_pos])
+                else:
+                    max_idx = bisect_left(_lists[max_pos], maximum)
+
+        return self._islice(min_pos, min_idx, max_pos, max_idx, reverse)
 
     def __len__(self):
         """Return the number of elements in the list."""
@@ -794,10 +945,6 @@ class SortedList(MutableSequence):
 
         return self._loc(pos, idx)
 
-    def bisect(self, val):
-        """Same as bisect_left."""
-        return self.bisect_left(val)
-
     def bisect_right(self, val):
         """
         Same as *bisect_left*, but if *val* is already present, the insertion
@@ -817,6 +964,8 @@ class SortedList(MutableSequence):
 
         return self._loc(pos, idx)
 
+    bisect = bisect_right
+
     def count(self, val):
         """Return the number of occurrences of *val* in the list."""
         _maxes = self._maxes
@@ -848,11 +997,9 @@ class SortedList(MutableSequence):
 
     def copy(self):
         """Return a shallow copy of the sorted list."""
-        return SortedList(self, load=self._load)
+        return self.__class__(self, load=self._load)
 
-    def __copy__(self):
-        """Return a shallow copy of the sorted list."""
-        return self.copy()
+    __copy__ = copy
 
     def append(self, val):
         """
@@ -870,7 +1017,8 @@ class SortedList(MutableSequence):
         pos = len(_lists) - 1
 
         if val < _lists[pos][-1]:
-            raise ValueError
+            msg = '{0} not in sort order at index {1}'.format(repr(val), self._len)
+            raise ValueError(msg)
 
         _maxes[pos] = val
         _lists[pos].append(val)
@@ -889,14 +1037,14 @@ class SortedList(MutableSequence):
 
         if any(values[pos - 1] > values[pos]
                for pos in range(1, len(values))):
-            raise ValueError
+            raise ValueError('given sequence not in sort order')
 
         offset = 0
-        count = len(_lists) - 1
 
         if _maxes:
             if values[0] < _lists[-1][-1]:
-                raise ValueError
+                msg = '{0} not in sort order at index {1}'.format(repr(values[0]), self._len)
+                raise ValueError(msg)
 
             if len(_lists[-1]) < self._half:
                 _lists[-1].extend(values[:_load])
@@ -921,7 +1069,7 @@ class SortedList(MutableSequence):
                     child = (child - 1) >> 1
                 _index[0] += len_values
         else:
-            del self._index[:]
+            del _index[:]
 
         self._len += len(values)
 
@@ -946,9 +1094,10 @@ class SortedList(MutableSequence):
             self._len = 1
             return
 
-        if idx == 0:
+        if not idx:
             if val > _lists[0][0]:
-                raise ValueError
+                msg = '{0} not in sort order at index {1}'.format(repr(val), 0)
+                raise ValueError(msg)
             else:
                 _lists[0].insert(0, val)
                 self._expand(0)
@@ -958,7 +1107,8 @@ class SortedList(MutableSequence):
         if idx == _len:
             pos = len(_lists) - 1
             if _lists[pos][-1] > val:
-                raise ValueError
+                msg = '{0} not in sort order at index {1}'.format(repr(val), _len)
+                raise ValueError(msg)
             else:
                 _lists[pos].append(val)
                 _maxes[pos] = _lists[pos][-1]
@@ -980,7 +1130,8 @@ class SortedList(MutableSequence):
             self._expand(pos)
             self._len += 1
         else:
-            raise ValueError
+            msg = '{0} not in sort order at index {1}'.format(repr(val), idx)
+            raise ValueError(msg)
 
     def pop(self, idx=-1):
         """
@@ -989,7 +1140,7 @@ class SortedList(MutableSequence):
         as for slice indices.
         """
         if (idx < 0 and -idx > self._len) or (idx >= self._len):
-            raise IndexError
+            raise IndexError('pop index out of range')
 
         pos, idx = self._pos(idx)
         val = self._lists[pos][idx]
@@ -1007,16 +1158,16 @@ class SortedList(MutableSequence):
         _len, _maxes = self._len, self._maxes
 
         if not _maxes:
-            raise ValueError
+            raise ValueError('{0} is not in list'.format(repr(val)))
 
-        if start == None:
+        if start is None:
             start = 0
         if start < 0:
             start += _len
         if start < 0:
             start = 0
 
-        if stop == None:
+        if stop is None:
             stop = _len
         if stop < 0:
             stop += _len
@@ -1024,19 +1175,19 @@ class SortedList(MutableSequence):
             stop = _len
 
         if stop <= start:
-            raise ValueError
+            raise ValueError('{0} is not in list'.format(repr(val)))
 
         stop -= 1
         pos_left = bisect_left(_maxes, val)
 
         if pos_left == len(_maxes):
-            raise ValueError
+            raise ValueError('{0} is not in list'.format(repr(val)))
 
         _lists = self._lists
         idx_left = bisect_left(_lists[pos_left], val)
 
         if _lists[pos_left][idx_left] != val:
-            raise ValueError
+            raise ValueError('{0} is not in list'.format(repr(val)))
 
         left = self._loc(pos_left, idx_left)
 
@@ -1049,7 +1200,7 @@ class SortedList(MutableSequence):
             if start <= right:
                 return start
 
-        raise ValueError
+        raise ValueError('{0} is not in list'.format(repr(val)))
 
     def as_list(self):
         """Very efficiently convert the SortedList to a list."""
@@ -1063,7 +1214,7 @@ class SortedList(MutableSequence):
         """
         values = self.as_list()
         values.extend(that)
-        return SortedList(values)
+        return self.__class__(values, load=self._load)
 
     def __iadd__(self, that):
         """
@@ -1079,7 +1230,7 @@ class SortedList(MutableSequence):
         in SortedList.
         """
         values = self.as_list() * that
-        return SortedList(values)
+        return self.__class__(values, load=self._load)
 
     def __imul__(self, that):
         """
@@ -1092,39 +1243,44 @@ class SortedList(MutableSequence):
         return self
 
     def __eq__(self, that):
-        """Compare two iterables for equality."""
+        """Compare two Sequences for equality."""
         return ((self._len == len(that))
                 and all(lhs == rhs for lhs, rhs in zip(self, that)))
 
     def __ne__(self, that):
-        """Compare two iterables for inequality."""
+        """Compare two Sequences for inequality."""
         return ((self._len != len(that))
                 or any(lhs != rhs for lhs, rhs in zip(self, that)))
 
     def __lt__(self, that):
-        """Compare two iterables for less than."""
+        """Compare two Sequences for less than."""
         return ((self._len <= len(that))
                 and all(lhs < rhs for lhs, rhs in zip(self, that)))
 
     def __le__(self, that):
-        """Compare two iterables for less than equal."""
+        """Compare two Sequences for less than equal."""
         return ((self._len <= len(that))
                 and all(lhs <= rhs for lhs, rhs in zip(self, that)))
 
     def __gt__(self, that):
-        """Compare two iterables for greater than."""
+        """Compare two Sequences for greater than."""
         return ((self._len >= len(that))
                 and all(lhs > rhs for lhs, rhs in zip(self, that)))
 
     def __ge__(self, that):
-        """Compare two iterables for greater than equal."""
+        """Compare two Sequences for greater than equal."""
         return ((self._len >= len(that))
                 and all(lhs >= rhs for lhs, rhs in zip(self, that)))
 
     @recursive_repr
     def __repr__(self):
         """Return string representation of SortedList."""
-        return '{0}({1})'.format(self.__class__.__name__, repr(self.as_list()))
+        temp = '{0}({1}, load={2})'
+        return temp.format(
+            self.__class__.__name__,
+            repr(list(self)),
+            repr(self._load)
+        )
 
     def _check(self):
         try:
@@ -1181,25 +1337,38 @@ class SortedList(MutableSequence):
             if len(self._index):
                 assert len(self._index) == self._offset + len(self._lists)
                 assert self._len == self._index[0]
-                assert all(self._index[self._offset + pos] == len(self._lists[pos])
+
+                def test_offset_pos(pos):
+                    from_index = self._index[self._offset + pos]
+                    return from_index == len(self._lists[pos])
+
+                assert all(test_offset_pos(pos)
                            for pos in range(len(self._lists)))
+
                 for pos in range(self._offset):
                     child = (pos << 1) + 1
-                    if self._index[pos] == 0:
-                        assert child >= len(self._index)
+                    if child >= len(self._index):
+                        assert self._index[pos] == 0
                     elif child + 1 == len(self._index):
                         assert self._index[pos] == self._index[child]
                     else:
-                        assert self._index[pos] == self._index[child] + self._index[child + 1]
+                        child_sum = self._index[child] + self._index[child + 1]
+                        assert self._index[pos] == child_sum
 
         except:
-            import sys, traceback
+            import sys
+            import traceback
 
             traceback.print_exc(file=sys.stdout)
 
-            print(self._len, self._load, self._half, self._twice)
-            print(self._index)
-            print(self._maxes)
-            print(self._lists)
+            print('len', self._len)
+            print('load', self._load, self._half, self._twice)
+            print('offset', self._offset)
+            print('len_index', len(self._index))
+            print('index', self._index)
+            print('len_maxes', len(self._maxes))
+            print('maxes', self._maxes)
+            print('len_lists', len(self._lists))
+            print('lists', self._lists)
 
             raise
